# -*- coding: utf-8 -*-
"""
Created on Tue Sep 13 13:44:52 2016

@author: albert.kurucz
"""
from __future__ import absolute_import
# import sys
import serial
import threading
import logging
import argparse
from tempfile import NamedTemporaryFile
from time import time
from time import sleep
# from binascii import hexlify
import serial.tools.list_ports as lp
from intelhex import IntelHex
from intelhex import NotEnoughDataError
import bz2
from snaplib import ppcrypto
try:
    from .snapavr_version import __version__
    assert __version__
except (ImportError, ValueError):
    __version__ = "UNKNOWN"
try:
    from .snapavr_version import __revision__
    assert __revision__
except (ImportError, ValueError):
    __revision__ = "UNKNOWN"

default_timeout = 30.0   # 30 seconds


class Timeout(Exception):
    pass


class AddressResponseNotReceived(Exception):
    pass


class ExitResponseNotReceived(Exception):
    pass


class InvalidBlockSizeResponse(Exception):
    pass


class SerialPortIsNotFound(Exception):
    pass


class TooManySerialPortsFound(Exception):
    pass


class ChecksumError(Exception):
    pass


def compute_checksum(data):
    computed = 0
    for char in data:
        computed += ord(char)
    computed &= 0xffff
    return computed


class AvrBootloader(serial.Serial):
    def __init__(self, serial_port):
        logging.info('opening serial port: %s', serial_port)
        self.waiting = True
        serial.Serial.__init__(
            self,
            serial_port,
            115200
        )
        self.setRTS(True)
        self.timeout = 0.1
        self.thread = threading.Thread(target=self.start)
        self.thread.start()

    def start(self):
        logging.info('starting...')
        self.setRTS(False)
        while self.waiting:
            chin = self.read(1)
            if chin == '\xf9':
                self.write('\xf6')
                self.flush()
                self.read_signature()

    def read_id(self):
        self.flushInput()
        self.write('I')
        self.flush()
        self.id_read = self.read(3)
        if len(self.id_read) == 3:
            logging.info('version: %d', ord(self.id_read[0]))

    def read_signature(self):
        self.flushInput()
        self.write('s')
        self.flush()
        self.signature_read = self.read(3)
        if len(self.signature_read) == 3:
            logging.info('sig1: %s', hex(ord(self.signature_read[2])))
            logging.info('sig2: %s', hex(ord(self.signature_read[1])))
            logging.info('sig3: %s', hex(ord(self.signature_read[0])))
            if self.signature_read == '\x04\x97\x1e':
                self.waiting = False

    def wait(self, timeout=default_timeout):
        end_time = time() + timeout
        while self.waiting:
            if time() > end_time:
                self.waiting = False
                self.thread.join()
                self.close()
                raise Timeout()
            sleep(0.1)

    def set_address(self, address):
        # logging.info('address: 0x%x', address)
        self.flushInput()
        self.write('A')
        self.write(chr(0xff & (address >> 9)))
        self.write(chr(0xff & (address >> 1)))
        self.flush()
        rec = self.read(1)
        if rec != '\r':
            raise AddressResponseNotReceived(
                'address: 0x%x, received: %s' % (address, repr(rec))
            )

    def get_block_size(self):
        self.flushInput()
        self.write('b')
        self.flush()
        self.blocksize_read = self.read(3)
        if len(self.blocksize_read) == 3:
            if self.blocksize_read[0] != 'Y':
                raise InvalidBlockSizeResponse()
            self.blocksize = (
                ord(self.blocksize_read[1]) << 8
            ) | ord(self.blocksize_read[2])
            logging.info('blocksize: %d', self.blocksize)

    def write_hex(self, intel_hex, memory_type='F'):
        # fist prepare the hex segments
        for segment_start, segment_end in intel_hex.segments():
            while segment_start % 256:
                segment_start -= 1
                try:
                    intel_hex.gets(segment_start, 1)
                except NotEnoughDataError:
                    intel_hex[segment_start] = 255
        # write the hex segments after prepared
        for segment_start, segment_end in intel_hex.segments():
            self.write_segment(
                intel_hex.gets(
                    segment_start,
                    segment_end - segment_start
                ),
                segment_start,
                memory_type,
            )

    def write_segment(self, data, segment_start, memory_type='F'):
        offset = 0
        segment_len = len(data)
        self.set_address(segment_start)
        while segment_len > 0:
            write_len = 256 if segment_len > 256 else segment_len
            try:
                self.write_block(
                    data[offset:offset + write_len],
                    memory_type
                )
            except ChecksumError:
                logging.error('checksum error at 0x%x', segment_start)
                raise
            offset += write_len
            segment_start += write_len
            segment_len -= write_len

    def write_block(self, data, memory_type='F'):
        data_len = len(data)
        self.flushInput()
        self.write('B')
        self.write(chr(0xff & (data_len >> 8)))
        self.write(chr(0xff & data_len))
        self.write(memory_type)
        self.write(data)
        self.flush()
        self.checksum_read = self.read(2)
        computed = compute_checksum(data)
        if len(self.checksum_read) == 2:
            self.checksum = (
                ord(self.checksum_read[0]) << 8
            ) | ord(self.checksum_read[1])
            if self.checksum != computed:
                logging.error(
                    'checksum: 0x%x (0x%x)' % (self.checksum, computed)
                )
                raise ChecksumError(
                    'checksum: 0x%x, computed: 0x%x' % (
                        self.checksum,
                        computed
                    )
                )

    def read_block(self, data_len, memory_type='F'):
        self.flushInput()
        self.write('g')
        self.write(chr(0xff & (data_len >> 8)))
        self.write(chr(0xff & data_len))
        self.write(memory_type)
        self.flush()
        # data_len = 2 * data_len
        data_read = self.read(data_len)
        actual_data_len = len(data_read)
        if actual_data_len != data_len:
            logging.error(
                'read length requested: %d, actual: %d',
                data_len,
                actual_data_len
            )
        return data_read

    def exit_bootloader(self):
        self.flushInput()
        self.write('E')
        self.flush()
        self.response_read = self.read(1)
        if self.response_read != '\r':
            raise ExitResponseNotReceived()


def snap_avr(
        serial_port=None,
        usb_vid=None,
        usb_pid=None,
        usb_serial_number=None,
        usb_location=None,
        timeout=default_timeout,
        do_not_exit_bootloader=False,
        sfi_file=None,
        hex_file=None,
        spy_file=None,
        magic=None,
):
    start_time = time()
    bl_time = None
    logging.info("timeout: %.3f", timeout)
    logging.info("do_not_exit_bootloader: %s", do_not_exit_bootloader)
    if serial_port:
        logging.info("serial_port: %s", serial_port)
    else:
        if usb_vid:
            logging.info("usb_vid: %d", usb_vid)
        if usb_pid:
            logging.info("usb_pid: %d", usb_pid)
        if usb_serial_number:
            logging.info("usb_serial_number: %s", usb_serial_number)
        if usb_location:
            logging.info("usb_location: %s", usb_location)
        com_ports = lp.comports()
        my_com_ports = []
        for com_port in reversed(com_ports):
            logging.info(
                "%s %d %d %s %s",
                com_port.device,
                com_port.vid,
                com_port.pid,
                repr(com_port.serial_number),
                repr(com_port.location),
            )
            if usb_vid:
                if usb_vid != com_port.vid:
                    continue
            if usb_pid:
                if usb_pid != com_port.pid:
                    continue
            if usb_serial_number:
                if usb_serial_number != com_port.serial_number:
                    continue
            if usb_location:
                if usb_location != com_port.location:
                    continue
            my_com_ports.append(com_port)
        if len(my_com_ports) == 0:
            raise SerialPortIsNotFound()
        if len(my_com_ports) > 1:
            raise TooManySerialPortsFound()
        serial_port = my_com_ports[0].device
    try:
        bl = AvrBootloader(serial_port)
        bl.wait(timeout)
        bl_time = time()
        bl.read_id()
        if sfi_file is not None:
            with open(sfi_file, 'rb') as sfi_stream:
                sfi_text = sfi_stream.read()
                # sfi files are compressed
                hex_text = bz2.decompress(sfi_text)
            with NamedTemporaryFile() as temp:
                temp.write(hex_text)
                temp.seek(0)
                bl.write_hex(IntelHex(temp.name))
        if hex_file is not None:
            bl.write_hex(IntelHex(hex_file))
        if spy_file is not None:
            with open(spy_file, 'rb') as spy_stream:
                spy_content = spy_stream.read()
            algo = ppcrypto.ARC4("SynapseExport3.14159265359")
            bl.write_segment(
                bz2.decompress(algo.decrypt(spy_content)),
                0xF00,
            )
        if magic is not None:
            bl.write_segment(magic, 0x2F0)
        if (hex_file is None) and (spy_file is None) and (magic is None):
            # by default perform a magical full erase
            bl.write_segment("RMDEMGKBF", 0x2F0)
        if do_not_exit_bootloader is False:
            bl.exit_bootloader()
    finally:
        bl.close()
    finish_time = time()
    logging.info('total runtime: %.3f' % (finish_time - start_time))
    if bl_time is not None:
        logging.info('programming time: %.3f' % (finish_time - bl_time))


def snap_avr_main():
    logging.basicConfig(level=logging.INFO)
    logging.info('snapavr version: %s', __version__)
    logging.info('snapavr revision: %s', __revision__)
    parser = argparse.ArgumentParser()
    parser.add_argument("--serial_port", help="serial port")
    parser.add_argument("--usb_location", help="usb location")
    parser.add_argument("--usb_pid", type=int, help="usb product id")
    parser.add_argument("--usb_vid", type=int, help="usb vendor id")
    parser.add_argument("--usb_serial_number", help="usb serial number")
    parser.add_argument("--sfi_file", help="sfi file")
    parser.add_argument("--hex_file", help="hex file")
    parser.add_argument("--spy_file", help="spy file")
    parser.add_argument("--magic", help="magic")
    parser.add_argument(
        "--do_not_exit_bootloader",
        dest='do_not_exit_bootloader',
        action='store_true',
        help="do not exit bootloader",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=default_timeout,
        help="timeout to engage bootloder",
    )
    args = parser.parse_args()
    snap_avr(
        serial_port=args.serial_port,
        usb_vid=args.usb_vid,
        usb_pid=args.usb_pid,
        usb_serial_number=args.usb_serial_number,
        usb_location=args.usb_location,
        timeout=args.timeout,
        do_not_exit_bootloader=args.do_not_exit_bootloader,
        sfi_file=args.sfi_file,
        hex_file=args.hex_file,
        spy_file=args.spy_file,
        magic=args.magic,
    )
    logging.info('done')
